/*
 * Copyright 2014, 2016 Peter Olsson
 * 
 * This file is part of Brum Brum Rally.
 * 
 * Brum Brum Rally is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * Brum Brum Rally is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "Chat.h"
#include "GameState.h"
#include "Settings.h"
#include "StateManager.h"
#include "Font.h"

Chat::Chat()
:	needUpdate(false),
	gameState(0),
	enabled(true),
	specialMessageVisible(false)
{
	specialMessage.renderObject.moveTo(2,2);
	specialMessage.timestamp = 0;
}

void Chat::addMessage(const std::string& message, Color color)
{
	const Font& font = getFont(Settings::network.chatFont ? BIG_FONT : SMALL_FONT);
	int messageWidth = font.calculateTextWidth(message);
	if (messageWidth >= GAME_WIDTH)
	{
		std::size_t halfWidth = message.size() / 2;
		addMessage(message.substr(0, halfWidth) + "...", color);
		addMessage("..."+message.substr(halfWidth), color);
		return;
	}
	
	messages.push_back(ChatMessage());
	ChatMessage& cm = messages.back();
	
	cm.renderObject.setSurface(Surface(message, Settings::network.chatFont ? BIG_FONT : SMALL_FONT, color, TEXT_BORDER_COLOR));
	if (gameState && enabled)
	{
		if (specialMessageVisible)
		{
			gameState->removeRenderObject(specialMessage.renderObject);
			specialMessageVisible = false;
		}
		gameState->addRenderObject(cm.renderObject);
	}
	// The exact time when the message was written is not important.
	// What's important is that it's shown long enough on the screen.
	cm.timestamp = SDL_GetTicks();
	
	needUpdate = true;
	
	if (messages.size() > Settings::network.maxChatMessages)
	{
		if (gameState)
		{
			gameState->removeRenderObject(messages.front().renderObject);
		}
		messages.pop_front();
	}
}

void Chat::update()
{
	Uint32 now = SDL_GetTicks();
	if (Settings::network.chatMessageDuration > 0)
	{
		if (SDL_GetAppState() & SDL_APPINPUTFOCUS)
		{
			while (!messages.empty() && messages.front().timestamp + Settings::network.chatMessageDuration * 1000 < now)
			{
				if (gameState)
				{
					gameState->removeRenderObject(messages.front().renderObject);
				}
				messages.pop_front();
				needUpdate = true;
			}
		}
		else
		{
			// Don't let the chat messages time out when the window isn't focused.
			std::deque<ChatMessage>::iterator it;
			for (it = messages.begin(); it != messages.end(); ++it)
			{
				// Reset timeout.
				it->timestamp = now + Settings::network.chatMessageDuration * 1000;
			}
		}
	}
	
	if (specialMessageVisible && specialMessage.timestamp + 2500 < now)
	{
		if (gameState)
		{
			gameState->removeRenderObject(specialMessage.renderObject);
		}
		specialMessageVisible = false;
	}
	
	if (!needUpdate)
	{
		return;
	}
	
	int x = 2;
	int y = 2;
	std::deque<ChatMessage>::iterator it;
	for (it = messages.begin(); it != messages.end(); ++it)
	{
		it->renderObject.moveTo(x, y);
		y += Settings::network.chatFont ? 10 : 8;
	}
	
	needUpdate = false;
}

void Chat::setGameState(GameState* newGameState)
{
	std::deque<ChatMessage>::iterator it;
	for (it = messages.begin(); it != messages.end(); ++it)
	{
		if (gameState)
		{
			gameState->removeRenderObject(it->renderObject);
		}
		if (newGameState)
		{
			newGameState->addRenderObject(it->renderObject);
		}
	}
	
	gameState = newGameState;
}

void Chat::toggle()
{
	specialMessageVisible = false;
	if (gameState)
	{
		gameState->removeRenderObject(specialMessage.renderObject);
	}
	
	std::deque<ChatMessage>::iterator it;
	if (enabled)
	{
		if (gameState)
		{
			for (it = messages.begin(); it != messages.end(); ++it)
			{
				gameState->removeRenderObject(it->renderObject);
			}
			setSpecialMessage("Chat off");
		}
		enabled = false;
	}
	else
	{
		if (gameState)
		{
			// Avoid showing messages with less than a second left to timeout.
			if (Settings::network.chatMessageDuration > 0)
			{
				const int MINIMUM_SHOW_TIME = 1000;
				while (!messages.empty() && messages.front().timestamp + Settings::network.chatMessageDuration * 1000 - MINIMUM_SHOW_TIME < SDL_GetTicks())
				{
					messages.front().timestamp -= MINIMUM_SHOW_TIME;
				}
				update();
			}
			
			if (messages.empty())
			{
				setSpecialMessage("Chat on");
			}
			else
			{
				for (it = messages.begin(); it != messages.end(); ++it)
				{
					gameState->addRenderObject(it->renderObject);
				}
			}
		}
		enabled = true;
	}
}

void Chat::toggle(bool enable)
{
	if (enable != isEnabled())
	{
		toggle();
	}
}

bool Chat::isEnabled() const
{
	return enabled;
}

void Chat::setSpecialMessage(const std::string& message)
{
	specialMessage.renderObject.setSurface(Surface(message, Settings::network.chatFont ? BIG_FONT : SMALL_FONT, CHAT_COLOR, TEXT_BORDER_COLOR));
	if (gameState)
	{
		gameState->addRenderObject(specialMessage.renderObject);
	}
	specialMessage.timestamp = SDL_GetTicks();
	specialMessageVisible = true;
}

void Chat::clear()
{
	setGameState(0);
	messages.clear();
}
